from __future__ import print_function

import argparse
import logging
import os
import shlex
import shutil
import sys
import traceback

from . import config
from . import __version__ as version

from PyInstaller.__main__ import run as run_pyinstaller


logger = logging.getLogger(__name__)


def __get_pyinstaller_argument_parser():
    from PyInstaller.building.makespec import __add_options as add_makespec_options
    from PyInstaller.building.build_main import __add_options as add_build_options
    from PyInstaller.log import __add_options as add_log_options

    parser = argparse.ArgumentParser()

    add_makespec_options(parser)
    add_build_options(parser)
    add_log_options(parser)

    parser.add_argument(
        'filenames', metavar='scriptname', nargs='+',
        help=("name of scriptfiles to be processed or "
              "exactly one .spec-file. If a .spec-file is "
              "specified, most options are unnecessary "
              "and are ignored.")
    )  # From PyInstaller.__main__.run

    return parser


def get_pyinstaller_options():
    parser = __get_pyinstaller_argument_parser()

    options = []
    for action in parser._actions:
        # Clean out what we can't send over to the ui
        # Here is what we currently have: https://github.com/python/cpython/blob/master/Lib/argparse.py#L771
        del action.container
        options.append(action)

    return [o.__dict__ for o in options]


def will_packaging_overwrite_existing(file_path, one_file, output_folder):
    """ Checks if there is a possibility of a previous output being overwritten. """
    if not os.path.exists(output_folder):
        return False
    no_extension = '.'.join(os.path.basename(file_path).split('.')[:-1])
    if one_file:
        if no_extension + '.exe' in os.listdir(output_folder):
            return True
    else:
        if no_extension in os.listdir(output_folder):
            return True
    return False


def __move_package(src, dst):
    """ Move the output package to the desired path (default is output/ - set in script.js) """
    # Make sure the destination exists
    if not os.path.exists(dst):
        os.makedirs(dst)

    # Move all files/folders in dist/
    for file_or_folder in os.listdir(src):
        _dst = os.path.join(dst, file_or_folder)
        # If this already exists in the destination, delete it
        if os.path.exists(_dst):
            if os.path.isfile(_dst):
                os.remove(_dst)
            else:
                shutil.rmtree(_dst)
        # Move file
        shutil.move(os.path.join(src, file_or_folder), dst)


def package(pyinstaller_command, options):
    """
    Call PyInstaller to package a script using provided arguments and options.
    :param pyinstaller_command: Command to supply to PyInstaller
    :param options: auto-py-to-exe specific options for setup and cleaning up
    :return: Whether packaging was successful
    """

    # Show current version
    logger.info("Running auto-py-to-exe v" + version)

    # Notify the user of the workspace and setup building to it
    logger.info("Building directory: {}".format(config.temporary_directory))

    # Override arguments
    dist_path = os.path.join(config.temporary_directory, 'application')
    build_path = os.path.join(config.temporary_directory, 'build')
    extra_args = ['--distpath', dist_path] + ['--workpath', build_path] + ['--specpath', config.temporary_directory]

    logger.info('Provided command: {}'.format(pyinstaller_command))

    # Setup options
    increase_recursion_limit = options['increaseRecursionLimit']
    output_directory = os.path.abspath(options['outputDirectory'])

    if increase_recursion_limit:
        sys.setrecursionlimit(5000)
        logger.info("Recursion Limit is set to 5000")
    else:
        sys.setrecursionlimit(config.DEFAULT_RECURSION_LIMIT)

    # Run PyInstaller
    fail = False
    try:
        # Since we allow manual argument input, we cannot pass arguments to PyInstaller as a list as we can't
        # guarantee that the arguments will be parsed correctly. To get around this, we can set sys.argv here with our
        # command to trick PyInstaller to reading the command as if we are using the cli tool.
        sys.argv = shlex.split(pyinstaller_command) + extra_args  # Put command into sys.argv and extra args

        # Display the command we are using and leave a space to separate out PyInstallers logs
        logger.info('Executing: {}'.format(' '.join(sys.argv)))
        logger.info('')

        run_pyinstaller()
    except:
        fail = True
        logger.exception("An error occurred while packaging")

    # Move project if there was no failure
    logger.info("")
    if not fail:
        logger.info("Moving project to: {0}".format(output_directory))
        try:
            __move_package(dist_path, output_directory)
        except:
            logger.error("Failed to move project")
            logger.exception(traceback.format_exc())
    else:
        logger.info("Project output will not be moved to output folder")
        return False

    # Set complete
    return True



import argparse
import json
import logging
import os


def argparse_file_eists(file_path):
    """ Validates whether a file exists """
    if not os.path.isfile(file_path):
        raise argparse.ArgumentTypeError('File does not exist')

    return os.path.abspath(file_path)


def argparse_file_json(file_path):
    """ Validates whether a file contains JSON parsable by Python. Raises argparse.ArgumentTypeError if not. """
    if not os.path.isfile(file_path):
        raise argparse.ArgumentTypeError('Provided configuration file does not exist')

    try:
        with open(file_path, 'r') as file:
            data = json.load(file)
    except json.decoder.JSONDecodeError:
        raise argparse.ArgumentTypeError('Provided configuration file content is not json')
    except Exception as e:
        raise argparse.ArgumentTypeError('Cannot parse provided configuration file:\n' + str(e))

    return data


def argparse_logging_level(level):
    """ Validates that a string value is a valid logging level and returns the corresponding value. """
    if hasattr(logging, level.upper()):
        return level.upper()
    raise argparse.ArgumentTypeError('Invalid logging level: ' + str(level))



from __future__ import print_function
import io
import os
import platform
import socket
import sys

from eel import chrome
from PyInstaller import __version__ as pyinstaller_version


class ForwardToFunctionStream(io.TextIOBase):
    def __init__(self, output_function=print):
        self.output_function = output_function

    def write(self, string):
        self.output_function(string)
        return len(string)


def can_use_chrome():
    """ Identify if Chrome is available for Eel to use """
    chrome_instance_path = chrome.find_path()
    return chrome_instance_path is not None and os.path.exists(chrome_instance_path)


def open_output_folder(folder):
    """ Open a folder in the local file explorer """
    folder_directory = os.path.abspath(folder)
    if platform.system() == 'Windows':
        os.startfile(folder_directory, 'explore')
    elif platform.system() == 'Linux':
        os.system('xdg-open "' + folder_directory + '"')
    elif platform.system() == 'Darwin':
        os.system('open "' + folder_directory + '"')
    else:
        return False
    return True


def get_warnings():
    warnings = []

    try:
        # Make sure PyInstaller 3.4 or above is being used with Python 3.7
        if sys.version_info >= (3, 7) and float(pyinstaller_version) < 3.4:
            message = 'You will need PyInstaller 3.4 or above to use this tool with Python 3.7'
            message += '\nPlease upgrade PyInstaller: python -m pip install pyinstaller --upgrade'
            warnings.append({
                'message': message,
                'link': None
            })
    except ValueError:
        pass  # Dev branches will have pyinstaller_version as a string in the form X.Y.devZ+HASH. Ignore it if this is the case.

    try:
        if sys.version_info.major == 3 and (sys.version_info.minor == 8 or sys.version_info.minor == 9) and float(pyinstaller_version) < 4.1:
            message = 'PyInstaller 4.0 and below do not officially support Python 3.8 and 3.9 - you are currently using PyInstaller {pyinstaller_version}'.format(pyinstaller_version=pyinstaller_version)
            message += '\nIt is highly recommended to update your version of PyInstaller using: python -m pip install pyinstaller --upgrade'
            warnings.append({
                'message': message,
                'link': None
            })
    except ValueError:
        pass  # Dev branches will have pyinstaller_version as a string in the form X.Y.devZ+HASH. Ignore it if this is the case.

    return warnings


def get_port():
    """ Get an available port by starting a new server, stopping and and returning the port """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(('localhost', 0))
    port = sock.getsockname()[1]
    sock.close()
    return port



import platform
import sys
try:
    from tkinter import Tk
except ImportError:
    try:
        from Tkinter import Tk
    except ImportError:
        # If no versions of tkinter exist (most likely linux) provide a message
        if sys.version_info.major < 3:
            print("Error: Tkinter not found")
            print('For linux, you can install Tkinter by executing: "sudo apt-get install python-tk"')
            sys.exit(1)
        else:
            print("Error: tkinter not found")
            print('For linux, you can install tkinter by executing: "sudo apt-get install python3-tk"')
            sys.exit(1)
try:
    from tkinter.filedialog import askopenfilename, askdirectory, askopenfilenames, asksaveasfilename
except ImportError:
    from tkFileDialog import askopenfilename, askdirectory, askopenfilenames, asksaveasfilename


def ask_file(file_type):
    """ Ask the user to select a file """
    root = Tk()
    root.withdraw()
    root.wm_attributes('-topmost', 1)
    if (file_type is None) or (platform.system() == "Darwin"):
        file_path = askopenfilename(parent=root)
    else:
        if file_type == 'python':
            file_types = [('Python files', '*.py;*.pyw'), ('All files', '*')]
        elif file_type == 'icon':
            file_types = [('Icon files', '*.ico'), ('All files', '*')]
        elif file_type == 'json':
            file_types = [('JSON Files', '*.json'), ('All files', '*')]
        else:
            file_types = [('All files', '*')]
        file_path = askopenfilename(parent=root, filetypes=file_types)
    root.update()

    """
    File dialog has 3 types of responses:
     - An empty tuple '()', when no files/folders were selected;
     - A single string '<filepath>', when a single file/folder was selected;
     - A tuple of strings '(<filepath1, ..., filepathN>)', when a multiple files/folders were selected;

    So bool() operator returns following values for the cases:
     - bool(()) -> False;
     - bool('<filepath>') -> True;
     - bool(('<filepath1>', ..., '<filepathN>')) -> True;
    """
    return file_path if bool(file_path) else None


def ask_files():
    """ Ask the user to select one or more files """
    root = Tk()
    root.withdraw()
    root.wm_attributes('-topmost', 1)
    file_paths = askopenfilenames(parent=root)
    root.update()

    """
    File dialog has 3 types of responses:
     - An empty tuple '()', when no files/folders were selected;
     - A single string '<filepath>', when a single file/folder was selected;
     - A tuple of strings '(<filepath1, ..., filepathN>)', when a multiple files/folders were selected;

    So bool() operator returns following values for the cases:
     - bool(()) -> False;
     - bool('<filepath>') -> True;
     - bool(('<filepath1>', ..., '<filepathN>')) -> True;
    """
    return file_paths if bool(file_paths) else None


def ask_folder():
    """ Ask the user to select a folder """
    root = Tk()
    root.withdraw()
    root.wm_attributes('-topmost', 1)
    folder = askdirectory(parent=root)
    root.update()

    """
    File dialog has 3 types of responses:
     - An empty tuple '()', when no files/folders were selected;
     - A single string '<filepath>', when a single file/folder was selected;
     - A tuple of strings '(<filepath1, ..., filepathN>)', when a multiple files/folders were selected;

    So bool() operator returns following values for the cases:
     - bool(()) -> False;
     - bool('<filepath>') -> True;
     - bool(('<filepath1>', ..., '<filepathN>')) -> True;
    """
    return folder if bool(folder) else None


def ask_file_save_location(file_type):
    """ Ask the user where to save a file """
    root = Tk()
    root.withdraw()
    root.wm_attributes('-topmost', 1)

    if (file_type is None) or (platform.system() == "Darwin"):
        file_path = asksaveasfilename(parent=root)
    else:
        if file_type == 'json':
            file_types = [('JSON Files', '*.json'), ('All files', '*')]
        else:
            file_types = [('All files', '*')]
        file_path = asksaveasfilename(parent=root, filetypes=file_types)
    root.update()

    """
    File dialog has 3 types of responses:
     - An empty tuple '()', when no files/folders were selected;
     - A single string '<filepath>', when a single file/folder was selected;
     - A tuple of strings '(<filepath1, ..., filepathN>)', when a multiple files/folders were selected;

    So bool() operator returns following values for the cases:
     - bool(()) -> False;
     - bool('<filepath>') -> True;
     - bool(('<filepath1>', ..., '<filepathN>')) -> True;
    """
    if bool(file_path):
        if file_type == 'json':
            return file_path if file_path.endswith('.json') else file_path + '.json'
        else:
            return file_path
    else:
        return None
